@@ -22,9 +22,8 @@
 #include <map>
 #include <set>
 #include <string>
-#include <boost/utility/string_view.hpp>
+#include <string_view>
 
-#include "common/DecayCounter.h"
 #include "common/bloom_filter.hpp"
 #include "common/config.h"
 #include "include/buffer_fwd.h"
@@ -33,6 +32,10 @@
 
 #include "CInode.h"
 #include "MDSCacheObject.h"
+#include "MDSContext.h"
+#include "cephfs_features.h"
+#include "SessionMap.h"
+#include "messages/MClientReply.h"
 
 class CDentry;
 class MDCache;
@@ -41,6 +44,9 @@ struct ObjectOperation;
 
 ostream& operator<<(ostream& out, const class CDir& dir);
 class CDir : public MDSCacheObject, public Counter<CDir> {
+  using time = ceph::coarse_mono_time;
+  using clock = ceph::coarse_mono_clock;
+
   friend ostream& operator<<(ostream& out, const class CDir& dir);
 
 public:
@@ -56,7 +62,7 @@ public:
   static const int PIN_EXPORTBOUND = 10;
   static const int PIN_STICKY =      11;
   static const int PIN_SUBTREETEMP = 12;  // used by MDCache::trim_non_auth()
-  const char *pin_name(int p) const override {
+  std::string_view pin_name(int p) const override {
     switch (p) {
     case PIN_DNWAITER: return "dnwaiter";
     case PIN_INOWAITER: return "inowaiter";
@@ -73,25 +79,26 @@ public:
   }
 
   // -- state --
-  static const unsigned STATE_COMPLETE =      (1<< 1);   // the complete contents are in cache
-  static const unsigned STATE_FROZENTREE =    (1<< 2);   // root of tree (bounded by exports)
-  static const unsigned STATE_FREEZINGTREE =  (1<< 3);   // in process of freezing 
-  static const unsigned STATE_FROZENDIR =     (1<< 4);
-  static const unsigned STATE_FREEZINGDIR =   (1<< 5);
-  static const unsigned STATE_COMMITTING =    (1<< 6);   // mid-commit
-  static const unsigned STATE_FETCHING =      (1<< 7);   // currenting fetching
-  static const unsigned STATE_CREATING =      (1<< 8);
-  static const unsigned STATE_IMPORTBOUND =   (1<<10);
-  static const unsigned STATE_EXPORTBOUND =   (1<<11);
-  static const unsigned STATE_EXPORTING =     (1<<12);
-  static const unsigned STATE_IMPORTING =     (1<<13);
-  static const unsigned STATE_FRAGMENTING =   (1<<14);
-  static const unsigned STATE_STICKY =        (1<<15);  // sticky pin due to inode stickydirs
-  static const unsigned STATE_DNPINNEDFRAG =  (1<<16);  // dir is refragmenting
-  static const unsigned STATE_ASSIMRSTAT =    (1<<17);  // assimilating inode->frag rstats
-  static const unsigned STATE_DIRTYDFT =      (1<<18);  // dirty dirfragtree
-  static const unsigned STATE_BADFRAG =       (1<<19);  // bad dirfrag
-  static const unsigned STATE_AUXSUBTREE =    (1<<20);  // no subtree merge
+  static const unsigned STATE_COMPLETE =      (1<< 0);   // the complete contents are in cache
+  static const unsigned STATE_FROZENTREE =    (1<< 1);   // root of tree (bounded by exports)
+  static const unsigned STATE_FREEZINGTREE =  (1<< 2);   // in process of freezing
+  static const unsigned STATE_FROZENDIR =     (1<< 3);
+  static const unsigned STATE_FREEZINGDIR =   (1<< 4);
+  static const unsigned STATE_COMMITTING =    (1<< 5);   // mid-commit
+  static const unsigned STATE_FETCHING =      (1<< 6);   // currenting fetching
+  static const unsigned STATE_CREATING =      (1<< 7);
+  static const unsigned STATE_IMPORTBOUND =   (1<< 8);
+  static const unsigned STATE_EXPORTBOUND =   (1<< 9);
+  static const unsigned STATE_EXPORTING =     (1<<10);
+  static const unsigned STATE_IMPORTING =     (1<<11);
+  static const unsigned STATE_FRAGMENTING =   (1<<12);
+  static const unsigned STATE_STICKY =        (1<<13);  // sticky pin due to inode stickydirs
+  static const unsigned STATE_DNPINNEDFRAG =  (1<<14);  // dir is refragmenting
+  static const unsigned STATE_ASSIMRSTAT =    (1<<15);  // assimilating inode->frag rstats
+  static const unsigned STATE_DIRTYDFT =      (1<<16);  // dirty dirfragtree
+  static const unsigned STATE_BADFRAG =       (1<<17);  // bad dirfrag
+  static const unsigned STATE_TRACKEDBYOFT =  (1<<18);  // tracked by open file table
+  static const unsigned STATE_AUXSUBTREE =    (1<<19);  // no subtree merge
 
   // common states
   static const unsigned STATE_CLEAN =  0;
@@ -102,18 +109,22 @@ public:
   (STATE_COMPLETE|STATE_DIRTY|STATE_DIRTYDFT|STATE_BADFRAG);
   static const unsigned MASK_STATE_IMPORT_KEPT = 
   (						  
-   STATE_IMPORTING
-   |STATE_IMPORTBOUND|STATE_EXPORTBOUND
-   |STATE_FROZENTREE
-   |STATE_STICKY);
+   STATE_IMPORTING |
+   STATE_IMPORTBOUND |
+   STATE_EXPORTBOUND |
+   STATE_FROZENTREE |
+   STATE_STICKY |
+   STATE_TRACKEDBYOFT);
   static const unsigned MASK_STATE_EXPORT_KEPT = 
-  (STATE_EXPORTING
-   |STATE_IMPORTBOUND|STATE_EXPORTBOUND
-   |STATE_FROZENTREE
-   |STATE_FROZENDIR
-   |STATE_STICKY);
+  (STATE_EXPORTING |
+   STATE_IMPORTBOUND |
+   STATE_EXPORTBOUND |
+   STATE_FROZENTREE |
+   STATE_FROZENDIR |
+   STATE_STICKY |
+   STATE_TRACKEDBYOFT);
   static const unsigned MASK_STATE_FRAGMENT_KEPT = 
-  (STATE_DIRTY|
+  (STATE_DIRTY |
    STATE_EXPORTBOUND |
    STATE_IMPORTBOUND |
    STATE_AUXSUBTREE |
@@ -137,11 +148,20 @@ public:
   static const int WAIT_DNLOCK_OFFSET = 4;
 
   static const uint64_t WAIT_ANY_MASK = (uint64_t)(-1);
-  static const uint64_t WAIT_ATFREEZEROOT = (WAIT_UNFREEZE);
   static const uint64_t WAIT_ATSUBTREEROOT = (WAIT_SINGLEAUTH);
 
-
-
+  // -- dump flags --
+  static const int DUMP_PATH             = (1 << 0);
+  static const int DUMP_DIRFRAG          = (1 << 1);
+  static const int DUMP_SNAPID_FIRST     = (1 << 2);
+  static const int DUMP_VERSIONS         = (1 << 3);
+  static const int DUMP_REP              = (1 << 4);
+  static const int DUMP_DIR_AUTH         = (1 << 5);
+  static const int DUMP_STATES           = (1 << 6);
+  static const int DUMP_MDS_CACHE_OBJECT = (1 << 7);
+  static const int DUMP_ITEMS            = (1 << 8);
+  static const int DUMP_ALL              = (-1);
+  static const int DUMP_DEFAULT          = DUMP_ALL & (~DUMP_ITEMS); 
 
  public:
   // context
@@ -186,7 +206,7 @@ public:
 public:
   version_t get_version() const { return fnode.version; }
   void set_version(version_t v) { 
-    assert(projected_fnode.empty());
+    ceph_assert(projected_fnode.empty());
     projected_version = fnode.version = v; 
   }
   version_t get_projected_version() const { return projected_version; }
@@ -286,12 +306,12 @@ public:
    * ENOENT: there are no remaining dentries to scrub
    * <0: There was an unexpected error
    *
-   * @param cb An MDSInternalContext which will be activated only if
+   * @param cb An MDSContext which will be activated only if
    *   we return EAGAIN via rcode, or else ignored
    * @param dnout CDentry * which you should next scrub, or NULL
    * @returns a value as described above
    */
-  int scrub_dentry_next(MDSInternalContext *cb, CDentry **dnout);
+  int scrub_dentry_next(MDSContext *cb, CDentry **dnout);
   /**
    * Get the currently scrubbing dentries. When returned, the passed-in
    * list will be filled with all CDentry * which have been returned
@@ -329,7 +349,7 @@ private:
    * next key to scrub and look it up (or fail!).
    */
   int _next_dentry_on_set(dentry_key_set &dns, bool missing_okay,
-                          MDSInternalContext *cb, CDentry **dnout);
+                          MDSContext *cb, CDentry **dnout);
 
 
 protected:
@@ -344,6 +364,8 @@ protected:
 
   int num_dirty;
 
+  int num_inodes_with_caps = 0;
+
   // state
   version_t committing_version;
   version_t committed_version;
@@ -355,7 +377,6 @@ protected:
   static int num_freezing_trees;
 
   int dir_auth_pins;
-  int request_pins;
 
   // cache control  (defined for authority; hints for replicas)
   __s32      dir_rep;
@@ -367,7 +388,7 @@ protected:
   dirfrag_load_vec_t pop_auth_subtree;
   dirfrag_load_vec_t pop_auth_subtree_nested;
  
-  utime_t last_popularity_sample;
+  time last_popularity_sample = clock::zero();
 
   load_spread_t pop_spread;
 
@@ -431,30 +452,29 @@ protected:
 
   void inc_num_dirty() { num_dirty++; }
   void dec_num_dirty() { 
-    assert(num_dirty > 0);
+    ceph_assert(num_dirty > 0);
     num_dirty--; 
   }
   int get_num_dirty() const {
     return num_dirty;
   }
 
+  void adjust_num_inodes_with_caps(int d);
+
   int64_t get_frag_size() const {
     return get_projected_fnode()->fragstat.size();
   }
 
   // -- dentries and inodes --
  public:
-  CDentry* lookup_exact_snap(boost::string_view dname, snapid_t last);
-  CDentry* lookup(boost::string_view n, snapid_t snap=CEPH_NOSNAP);
-  CDentry* lookup(const char *n, snapid_t snap=CEPH_NOSNAP) {
-    return lookup(boost::string_view(n), snap);
-  }
+  CDentry* lookup_exact_snap(std::string_view dname, snapid_t last);
+  CDentry* lookup(std::string_view n, snapid_t snap=CEPH_NOSNAP);
 
-  CDentry* add_null_dentry(boost::string_view dname,
+  CDentry* add_null_dentry(std::string_view dname,
 			   snapid_t first=2, snapid_t last=CEPH_NOSNAP);
-  CDentry* add_primary_dentry(boost::string_view dname, CInode *in,
+  CDentry* add_primary_dentry(std::string_view dname, CInode *in,
 			      snapid_t first=2, snapid_t last=CEPH_NOSNAP);
-  CDentry* add_remote_dentry(boost::string_view dname, inodeno_t ino, unsigned char d_type,
+  CDentry* add_remote_dentry(std::string_view dname, inodeno_t ino, unsigned char d_type,
 			     snapid_t first=2, snapid_t last=CEPH_NOSNAP);
   void remove_dentry( CDentry *dn );         // delete dentry
   void link_remote_inode( CDentry *dn, inodeno_t ino, unsigned char d_type);
@@ -464,7 +484,7 @@ protected:
   void try_remove_unlinked_dn(CDentry *dn);
 
   void add_to_bloom(CDentry *dn);
-  bool is_in_bloom(boost::string_view name);
+  bool is_in_bloom(std::string_view name);
   bool has_bloom() { return (bloom ? true : false); }
   void remove_bloom() {
     bloom.reset();
@@ -480,22 +500,22 @@ public:
 
 
 public:
-  void split(int bits, std::list<CDir*>& subs, list<MDSInternalContextBase*>& waiters, bool replay);
-  void merge(std::list<CDir*>& subs, std::list<MDSInternalContextBase*>& waiters, bool replay);
+  void split(int bits, std::list<CDir*>& subs, MDSContext::vec& waiters, bool replay);
+  void merge(std::list<CDir*>& subs, MDSContext::vec& waiters, bool replay);
 
   bool should_split() const {
-    return (int)get_frag_size() > g_conf->mds_bal_split_size;
+    return (int)get_frag_size() > g_conf()->mds_bal_split_size;
   }
   bool should_split_fast() const;
   bool should_merge() const {
-    return (int)get_frag_size() < g_conf->mds_bal_merge_size;
+    return (int)get_frag_size() < g_conf()->mds_bal_merge_size;
   }
 
 private:
   void prepare_new_fragment(bool replay);
-  void prepare_old_fragment(map<string_snap_t, std::list<MDSInternalContextBase*> >& dentry_waiters, bool replay);
+  void prepare_old_fragment(map<string_snap_t, MDSContext::vec >& dentry_waiters, bool replay);
   void steal_dentry(CDentry *dn);  // from another dir.  used by merge/split.
-  void finish_old_fragment(std::list<MDSInternalContextBase*>& waiters, bool replay);
+  void finish_old_fragment(MDSContext::vec& waiters, bool replay);
   void init_fragment_pins();
 
 
@@ -513,7 +533,7 @@ private:
  public:
   mds_authority_t authority() const override;
   mds_authority_t get_dir_auth() const { return dir_auth; }
-  void set_dir_auth(mds_authority_t a);
+  void set_dir_auth(const mds_authority_t &a);
   void set_dir_auth(mds_rank_t a) { set_dir_auth(mds_authority_t(a, CDIR_AUTH_UNKNOWN)); }
   bool is_ambiguous_dir_auth() const {
     return dir_auth.second != CDIR_AUTH_UNKNOWN;
@@ -534,49 +554,35 @@ private:
 
   // for giving to clients
   void get_dist_spec(std::set<mds_rank_t>& ls, mds_rank_t auth) {
-    if (is_rep()) {
+    if (is_auth()) {
       list_replicas(ls);
       if (!ls.empty()) 
 	ls.insert(auth);
     }
   }
-  void encode_dirstat(bufferlist& bl, mds_rank_t whoami) {
-    /*
-     * note: encoding matches struct ceph_client_reply_dirfrag
-     */
-    frag_t frag = get_frag();
-    mds_rank_t auth;
-    std::set<mds_rank_t> dist;
-    
-    auth = dir_auth.first;
-    if (is_auth()) 
-      get_dist_spec(dist, whoami);
 
-    ::encode(frag, bl);
-    ::encode(auth, bl);
-    ::encode(dist, bl);
-  }
+  static void encode_dirstat(bufferlist& bl, const session_info_t& info, const DirStat& ds);
 
   void _encode_base(bufferlist& bl) {
-    ::encode(first, bl);
-    ::encode(fnode, bl);
-    ::encode(dir_rep, bl);
-    ::encode(dir_rep_by, bl);
+    encode(first, bl);
+    encode(fnode, bl);
+    encode(dir_rep, bl);
+    encode(dir_rep_by, bl);
   }
-  void _decode_base(bufferlist::iterator& p) {
-    ::decode(first, p);
-    ::decode(fnode, p);
-    ::decode(dir_rep, p);
-    ::decode(dir_rep_by, p);
+  void _decode_base(bufferlist::const_iterator& p) {
+    decode(first, p);
+    decode(fnode, p);
+    decode(dir_rep, p);
+    decode(dir_rep_by, p);
   }
   void encode_replica(mds_rank_t who, bufferlist& bl) {
     __u32 nonce = add_replica(who);
-    ::encode(nonce, bl);
+    encode(nonce, bl);
     _encode_base(bl);
   }
-  void decode_replica(bufferlist::iterator& p) {
+  void decode_replica(bufferlist::const_iterator& p) {
     __u32 nonce;
-    ::decode(nonce, p);
+    decode(nonce, p);
     replica_nonce = nonce;
     _decode_base(p);
   }
@@ -599,34 +605,28 @@ private:
   object_t get_ondisk_object() { 
     return file_object_t(ino(), frag);
   }
-  void fetch(MDSInternalContextBase *c, bool ignore_authpinnability=false);
-  void fetch(MDSInternalContextBase *c, boost::string_view want_dn, bool ignore_authpinnability=false);
-  void fetch(MDSInternalContextBase *c, const std::set<dentry_key_t>& keys);
+  void fetch(MDSContext *c, bool ignore_authpinnability=false);
+  void fetch(MDSContext *c, std::string_view want_dn, bool ignore_authpinnability=false);
+  void fetch(MDSContext *c, const std::set<dentry_key_t>& keys);
 protected:
   mempool::mds_co::compact_set<mempool::mds_co::string> wanted_items;
 
-  void _omap_fetch(MDSInternalContextBase *fin, const std::set<dentry_key_t>& keys);
-  void _omap_fetch_more(
-    bufferlist& hdrbl, std::map<std::string, bufferlist>& omap,
-    MDSInternalContextBase *fin);
+  void _omap_fetch(MDSContext *fin, const std::set<dentry_key_t>& keys);
+  void _omap_fetch_more(version_t omap_version, bufferlist& hdrbl,
+                        map<string, bufferlist>& omap, MDSContext *fin);
   CDentry *_load_dentry(
-      boost::string_view key,
-      boost::string_view dname,
+      std::string_view key,
+      std::string_view dname,
       snapid_t last,
       bufferlist &bl,
       int pos,
       const std::set<snapid_t> *snaps,
       bool *force_dirty);
 
-  /**
-   * Mark this fragment as BADFRAG (common part of go_bad and go_bad_dentry)
-   */
-  void _go_bad();
-
   /**
    * Go bad due to a damaged dentry (register with damagetable and go BADFRAG)
    */
-  void go_bad_dentry(snapid_t last, boost::string_view dname);
+  void go_bad_dentry(snapid_t last, std::string_view dname);
 
   /**
    * Go bad due to a damaged header (register with damagetable and go BADFRAG)
@@ -637,7 +637,7 @@ protected:
 		     bool complete, int r);
 
   // -- commit --
-  mempool::mds_co::compact_map<version_t, mempool::mds_co::list<MDSInternalContextBase*> > waiting_for_commit;
+  mempool::mds_co::compact_map<version_t, MDSContext::vec_alloc<mempool::mds_co::pool_allocator> > waiting_for_commit;
   void _commit(version_t want, int op_prio);
   void _omap_commit(int op_prio);
   void _encode_dentry(CDentry *dn, bufferlist& bl, const std::set<snapid_t> *snaps);
@@ -647,7 +647,7 @@ public:
   void wait_for_commit(Context *c, version_t v=0);
 #endif
   void commit_to(version_t want);
-  void commit(version_t want, MDSInternalContextBase *c,
+  void commit(version_t want, MDSContext *c,
 	      bool ignore_authpinnability=false, int op_prio=-1);
 
   // -- dirtyness --
@@ -662,56 +662,58 @@ public:
   void first_get() override;
   void last_put() override;
 
-  void request_pin_get() {
-    if (request_pins == 0) get(PIN_REQUEST);
-    request_pins++;
-  }
-  void request_pin_put() {
-    request_pins--;
-    if (request_pins == 0) put(PIN_REQUEST);
-  }
-
   // -- waiters --
 protected:
-  mempool::mds_co::compact_map< string_snap_t, mempool::mds_co::list<MDSInternalContextBase*> > waiting_on_dentry; // FIXME string_snap_t not in mempool
+  mempool::mds_co::compact_map< string_snap_t, MDSContext::vec_alloc<mempool::mds_co::pool_allocator> > waiting_on_dentry; // FIXME string_snap_t not in mempool
 
 public:
-  bool is_waiting_for_dentry(boost::string_view dname, snapid_t snap) {
+  bool is_waiting_for_dentry(std::string_view dname, snapid_t snap) {
     return waiting_on_dentry.count(string_snap_t(dname, snap));
   }
-  void add_dentry_waiter(boost::string_view dentry, snapid_t snap, MDSInternalContextBase *c);
-  void take_dentry_waiting(boost::string_view dentry, snapid_t first, snapid_t last, std::list<MDSInternalContextBase*>& ls);
-  void take_sub_waiting(std::list<MDSInternalContextBase*>& ls);  // dentry or ino
+  void add_dentry_waiter(std::string_view dentry, snapid_t snap, MDSContext *c);
+  void take_dentry_waiting(std::string_view dentry, snapid_t first, snapid_t last, MDSContext::vec& ls);
+  void take_sub_waiting(MDSContext::vec& ls);  // dentry or ino
 
-  void add_waiter(uint64_t mask, MDSInternalContextBase *c) override;
-  void take_waiting(uint64_t mask, std::list<MDSInternalContextBase*>& ls) override;  // may include dentry waiters
+  void add_waiter(uint64_t mask, MDSContext *c) override;
+  void take_waiting(uint64_t mask, MDSContext::vec& ls) override;  // may include dentry waiters
   void finish_waiting(uint64_t mask, int result = 0);    // ditto
   
 
   // -- import/export --
   void encode_export(bufferlist& bl);
-  void finish_export(utime_t now);
+  void finish_export();
   void abort_export() {
     put(PIN_TEMPEXPORTING);
   }
-  void decode_import(bufferlist::iterator& blp, utime_t now, LogSegment *ls);
+  void decode_import(bufferlist::const_iterator& blp, LogSegment *ls);
+  void abort_import();
 
   // -- auth pins --
   bool can_auth_pin(int *err_ret=nullptr) const override;
-  int get_cum_auth_pins() const { return auth_pins + nested_auth_pins; }
   int get_auth_pins() const { return auth_pins; }
-  int get_nested_auth_pins() const { return nested_auth_pins; }
   int get_dir_auth_pins() const { return dir_auth_pins; }
   void auth_pin(void *who) override;
   void auth_unpin(void *who) override;
 
-  void adjust_nested_auth_pins(int inc, int dirinc, void *by);
+  void adjust_nested_auth_pins(int dirinc, void *by);
   void verify_fragstat();
 
   // -- freezing --
+  struct freeze_tree_state_t {
+    CDir *dir; // freezing/frozen tree root
+    int auth_pins = 0;
+    bool frozen = false;
+    freeze_tree_state_t(CDir *d) : dir(d) {}
+  };
+  // all dirfrags within freezing/frozen tree reference the 'state'
+  std::shared_ptr<freeze_tree_state_t> freeze_tree_state;
+
+  void _walk_tree(std::function<bool(CDir*)> cb);
+
   bool freeze_tree();
   void _freeze_tree();
   void unfreeze_tree();
+  void adjust_freeze_after_rename(CDir *dir);
 
   bool freeze_dir();
   void _freeze_dir();
@@ -719,7 +721,14 @@ public:
 
   void maybe_finish_freeze();
 
-  pair<bool,bool> is_freezing_or_frozen_tree() const;
+  pair<bool,bool> is_freezing_or_frozen_tree() const {
+    if (freeze_tree_state) {
+      if (freeze_tree_state->frozen)
+	return make_pair(false, true);
+      return make_pair(true, false);
+    }
+    return make_pair(false, false);
+  }
 
   bool is_freezing() const override { return is_freezing_dir() || is_freezing_tree(); }
   bool is_freezing_tree() const {
@@ -729,7 +738,6 @@ public:
   }
   bool is_freezing_tree_root() const { return state & STATE_FREEZINGTREE; }
   bool is_freezing_dir() const { return state & STATE_FREEZINGDIR; }
-  CDir *get_freezing_tree_root();
 
   bool is_frozen() const override { return is_frozen_dir() || is_frozen_tree(); }
   bool is_frozen_tree() const {
@@ -739,11 +747,11 @@ public:
   }
   bool is_frozen_tree_root() const { return state & STATE_FROZENTREE; }
   bool is_frozen_dir() const { return state & STATE_FROZENDIR; }
-  CDir *get_frozen_tree_root();
-  
+
   bool is_freezeable(bool freezing=false) const {
     // no nested auth pins.
-    if ((auth_pins-freezing) > 0 || nested_auth_pins > 0) 
+    if (auth_pins - (freezing ? 1 : 0) > 0 ||
+	(freeze_tree_state && freeze_tree_state->auth_pins != auth_pins))
       return false;
 
     // inode must not be frozen.
@@ -752,8 +760,9 @@ public:
 
     return true;
   }
+
   bool is_freezeable_dir(bool freezing=false) const {
-    if ((auth_pins-freezing) > 0 || dir_auth_pins > 0) 
+    if ((auth_pins - freezing) > 0 || dir_auth_pins > 0)
       return false;
 
     // if not subtree root, inode must not be frozen (tree--frozen_dir is okay).
@@ -765,8 +774,8 @@ public:
 
   ostream& print_db_line_prefix(ostream& out) override;
   void print(ostream& out) override;
-  void dump(Formatter *f) const;
-  void dump_load(Formatter *f, utime_t now, const DecayRate& rate);
+  void dump(Formatter *f, int flags = DUMP_DEFAULT) const;
+  void dump_load(Formatter *f);
 };
 
 #endif
